/*****************************************************
 * SAXPY - Scalar Alpha X Plus Y: y[i] = a*x[i] + y[i]
 * CUDA C
 *====================================================
 * Center for Computing and Communication (CCC)
 * RWTH Aachen University, Germany
 * Author: Sandra Wienke (wienke@rz.rwth-aachen.de)
 * Date: February 2011
 *****************************************************/

#include <stdio.h>
#include "realtime.h"

// Declaration of helper functions (see below for details)
void printParameter(unsigned int,float,int);
void printPerformanceMetrics(double, double, double, unsigned int);
float getGflops(double, unsigned int);
void saxpy_serial(unsigned int, float, float*, float*);
int checkResults(float*,float*,unsigned int);



// GPU kernel function
__global__ void saxpy_parallel(unsigned int n, float a, float *x, float *y)
{
    // TODO: Compute index i using thread and block indices


    if (i < n){
      y[i] = a*x[i] + y[i];
     }
}



int main(int argc, char* argv[])
{
  // For time measurements
  double tiStart1, tiStop1, tiStart2, tiStop2, tiStart3, tiStop3;
  double timeCPU, timeGPU, timeGPU_kernel;

  int gpuDev = -1; // Device no of GPU

  unsigned int n = 262144; // MAX 1677690 on Fermi (with 256 threads/block)
  float a = 2.0;	
  int blockDim = 256;
  if (argc > 1) {
    n = atoi(argv[1]);
  }
  if (argc > 2) {
    blockDim = atoi(argv[2]);
  }
  printParameter(n, a, blockDim);


  float* h_x, *h_y, *h_yy;	// Pointer to CPU memory
  h_x = (float*) malloc(n * sizeof(float)); 
  h_y = (float*) malloc(n * sizeof(float));
  h_yy = (float*) malloc(n * sizeof(float)); // just for comparison of results

  // Initialize h_x and h_y
  for(unsigned int i=0; i<n; ++i){
    h_x[i]=i;
    h_y[i]=h_yy[i]=5.0*i-1.0;
  }


  float *d_x, *d_y;	// Pointer to GPU memory


  // TODO: Allocate d_x and d_y on the device



  tiStart1 = GetRealTime();

  // TODO: Copy h_x into d_x, and h_y into d_y




  // TODO: Define 1D thread blocks of length blockDim


  dim3 blocksPerGrid((n%blockDim==0)? n/threadsPerBlock.x : n/threadsPerBlock.x +1);

  cudaThreadSynchronize();
  tiStart2 = GetRealTime();


  // TODO: Invoke parallel SAXPY kernel
 

  cudaThreadSynchronize();
  tiStop2 = GetRealTime();


  // TODO: Copy d_y into h_y
 

  tiStop1 = GetRealTime();

  cudaGetDevice(&gpuDev);
  printf("Running on device %d\n", gpuDev)

  // TODO: Free memory on device (d_x, d_y)
 
 


  // Compare CPU and GPU results
  tiStart3 = GetRealTime();
  saxpy_serial(n,a,h_x,h_yy);
  tiStop3 = GetRealTime();
  checkResults(h_yy, h_y, n);


  // Compute elapsed time
  timeCPU = tiStop3 - tiStart3;
  timeGPU = tiStop1 - tiStart1;
  timeGPU_kernel = tiStop2 - tiStart2;
  printPerformanceMetrics(timeCPU, timeGPU, timeGPU_kernel, n);


  free(h_x);
  free(h_y);
  free(h_yy);
  return 0;
}



// Get performance metric gflops (time in sec)
float getGflops (double time, unsigned int n) {
	float gf = 0;
	unsigned long int operations;

	operations = (unsigned long int) (2 * n);
	gf = operations / time;
	gf /= 1000; // Kilo
	gf /= 1000; // Mega
	gf /= 1000; // Giga
	return gf;
}

// Print vector size, scalar a and number threads per block
void printParameter(unsigned int n, float a, int blockDim) {
  printf("Vector size n:\t%d\n", n);
  printf("Scalar a:\t%f\n", a);
  printf("#Threads/block:\t%d\n\n", blockDim);
}

// Print runtime, GFlops and speedup
void printPerformanceMetrics(double timeCPU, double timeGPU, double timeGPU_kernel, unsigned int n) {
  printf("\t\t\t\t| Runtime [s]\t| GFlops\t| Speedup\t|\n");
  printf("=================================================================================\n");
  printf("CPU (serial)\t\t\t| %.10f\t| %f\t| %f\t|\n", timeCPU, getGflops(timeCPU,n), 1.f);
  printf("GPU (kernel)\t\t\t| %.10f\t| %f\t| %f\t|\n", timeGPU_kernel, getGflops(timeGPU_kernel,n), timeCPU / timeGPU_kernel);  
  printf("GPU (kernel + data transfer)\t| %.10f\t| %f\t| %f\t|\n", timeGPU, getGflops(timeGPU,n), timeCPU / timeGPU);
}

// Compute SAXPY on CPU
void saxpy_serial(unsigned int n, float a, float* x, float* y) {
  for(unsigned int i=0; i<n; ++i) {
    y[i] = a*x[i] + y[i];
  }
}

// Compare CPU and GPU results
int checkResults(float* cpu, float* gpu, unsigned int n) {
  float tolerance = 1e-6f;
  for(unsigned int i=0; i<n; ++i){
    if (abs(cpu[i] - gpu[i]) > tolerance){
      printf("Error check failed\n");
      return 1;
    }
  }
  printf("Error check passed\n");
  return 0;
}
