/*****************************************************
 * SAXPY - Scalar Alpha X Plus Y: y[i] = a*x[i] + y[i]
 * PGI Accelerator for C
 *====================================================
 * Center for Computing and Communication (CCC)
 * RWTH Aachen University, Germany
 * Author: Sandra Wienke (wienke@rz.rwth-aachen.de)
 * Date: February 2011
 *****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <accel.h>
#include "realtime.h"

// Declaration of helper functions (see below for details)
void printParameter(unsigned int,float);
void printPerformanceMetrics(double, double, unsigned int);
float getGflops(double, unsigned int);
void saxpy_serial(unsigned int, float, float*, float*);
int checkResults(float*,float*,unsigned int);

int main(int argc, char* argv[])
{
    // For time measurements
  double tiStart1, tiStop1, tiStart2, tiStop2, tiStart3, tiStop3;
  double timeCPU, timeGPU;

  unsigned int n = 262144;
  float a = 2.0;
  if (argc > 1) {
    n = atoi(argv[1]);
  }
  printParameter(n, a);
	
  float* restrict x;
  float* restrict y;
  float* restrict yy; // just for comparison of results
  x = (float*) malloc(n * sizeof(float)); 
  y = (float*) malloc(n * sizeof(float));
  yy = (float*) malloc(n * sizeof(float));

  // Initialize x and y
  for(int i=0; i<n; ++i){
    x[i]=i;
    y[i]=yy[i]=5.0*i-1.0;
  }

  // Initialize GPU device (for better performance afterwards)
  acc_init(acc_device_nvidia); 
    
  tiStart1 = GetRealTime();

  // TODO 1: Define accelerator region 
  // TODO 2: Add data clauses
  // TODO 3: Define loop scheduling
#pragma acc region for copy(y[0:n-1]), copyin(x[0:n-1]) parallel,vector(256)
  {
    for (unsigned int i=0; i  < n ; ++i) {
      y[i] = a*x[i] + y[i];
    }
  }
  
  tiStop1 = GetRealTime();

   // Compare CPU and GPU results
  tiStart2 = GetRealTime();
  saxpy_serial(n,a,x,yy);
  tiStop2 = GetRealTime();
  checkResults(yy, y, n);

  // Compute elapsed time
  timeCPU = tiStop2 - tiStart2;
  timeGPU = tiStop1 - tiStart1;
  printPerformanceMetrics(timeCPU, timeGPU, n);

  free(x);
  free(y);
  free(yy);
  return 0;
}


// Get performance metric gflops (time in sec)
float getGflops (double time, unsigned int n) {
	float gf = 0;
	unsigned long int operations;

	operations = (unsigned long int) (2 * n);
	gf = operations / time;
	gf /= 1000; // Kilo
	gf /= 1000; // Mega
	gf /= 1000; // Giga
	return gf;
}

// Print vector size and  scalar a
void printParameter(unsigned int n, float a) {
  printf("Vector size n:\t%d\n", n);
  printf("Scalar a:\t%f\n\n", a);
}

// Print runtime, GFlops and speedup
void printPerformanceMetrics(double timeCPU, double timeGPU, unsigned int n) {
  printf("\t\t\t\t| Runtime [s]\t| GFlops\t| Speedup\t|\n");
  printf("=================================================================================\n");
  printf("CPU (serial)\t\t\t| %.10f\t| %f\t| %f\t|\n", timeCPU, getGflops(timeCPU,n), 1.f);
  printf("GPU (kernel + data transfer)\t| %.10f\t| %f\t| %f\t|\n", timeGPU, getGflops(timeGPU,n), timeCPU / timeGPU);
}

// Compute SAXPY on CPU
void saxpy_serial(unsigned int n, float a, float* x, float* y) {
  for(unsigned int i=0; i<n; ++i) {
    y[i] = a*x[i] + y[i];
  }
}

// Compare CPU and GPU results
int checkResults(float* cpu, float* gpu, unsigned int n) {
  float tolerance = 1e-6f;
  for(unsigned int i=0; i<n; ++i){
    if (abs(cpu[i] - gpu[i]) > tolerance){
      printf("Error check failed\n");
      return 1;
    }
  }
  printf("Error check passed\n");
  return 0;
}
